﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace JonathanHaas.ActiveVB.WB3 {
    public class PathFinder {
        private double _bestPathLength = Double.MaxValue;

        public void FindBestPath(int firstTypeIndex, int lastTypeIndex, IList<ICollection<PathNode>> pathNodes) {
            var startNodes = pathNodes[firstTypeIndex];
            var endNodes = pathNodes[lastTypeIndex];
            List<DirectPath> paths = new List<DirectPath>(startNodes.Count * endNodes.Count);
            foreach (var startNode in startNodes) {
                foreach (var endNode in endNodes) {
                    paths.Add(new DirectPath(startNode, endNode));
                }
            }
            paths.Sort();
            foreach (var path in paths) {
                var g = FindGoodPath(path.startNode, path.endNode, firstTypeIndex + 1, lastTypeIndex - 1, pathNodes, 0);
                if (g != null) {
                    _bestPathLength = Math.Min(_bestPathLength, PathLength(g));
                    foreach (var node in g) {
                        System.Console.Write(node.Name);
                        System.Console.Write(", ");
                    }
                    System.Console.WriteLine(PathLength(g));
                }

                var p = FindBestPath(path.startNode, path.endNode, firstTypeIndex + 1, lastTypeIndex - 1, pathNodes, 0);
                if (p != null) {
                    _bestPathLength = Math.Min(_bestPathLength, PathLength(p));
                    foreach (var node in p) {
                        System.Console.Write(node.Name);
                        System.Console.Write(", ");
                    }
                    System.Console.WriteLine(PathLength(p));
                }
            }
        }

        private List<PathNode> FindBestPath(PathNode firstNode, PathNode lastNode, int firstTypeIndex, int lastTypeIndex, IList<ICollection<PathNode>> pathNodes, double extraLength) {
            if (lastTypeIndex < firstTypeIndex) {
                return new List<PathNode>(2) { firstNode, lastNode };
            }
            int midIndex = firstTypeIndex + (lastTypeIndex - firstTypeIndex) / 2;
            var paths = new List<ThreeNodePath>(pathNodes[midIndex].Count);
            foreach (var node in pathNodes[midIndex]) {
                var path = new ThreeNodePath(firstNode, node, lastNode);
                if (path.Length + extraLength < _bestPathLength)
                    paths.Add(path);
            }

            if (paths.Count == 0) {
                return null;
            }

            paths.Sort();

            List<PathNode> bestFirstPath = null;
            List<PathNode> bestLastPath = null;

            double bestLength = Double.MaxValue;

            if(extraLength == 0) System.Console.WriteLine(paths.Count);
            var i = 0;
            foreach (var path in paths) {
                if (extraLength == 0) System.Console.WriteLine(i);

                i++;
                var firstPath = FindBestPath(firstNode, path.MidNode, firstTypeIndex, midIndex - 1, pathNodes, path.LastSegmentLength + extraLength);
                var lastPath = FindBestPath(path.MidNode, lastNode, midIndex + 1, lastTypeIndex, pathNodes, path.FirstSegmentLength + extraLength);

                if (firstPath == null || lastPath == null) continue;

                var newLength = PathLength(firstPath) + PathLength(lastPath);
                if (newLength < bestLength) {
                    bestLength = newLength;
                    bestFirstPath = firstPath;
                    bestLastPath = lastPath;
                }

                if (path.Length > bestLength) {
                    break;
                }
            }


            if (bestFirstPath == null) {
                return null;
            }

            var result = new List<PathNode>(bestFirstPath.Count + bestLastPath.Count - 1);
            result.AddRange(bestFirstPath);
            result.AddRange(bestLastPath.GetRange(1, bestLastPath.Count - 1));
            return result;
        }


        private List<PathNode> FindGoodPath(PathNode firstNode, PathNode lastNode, int firstTypeIndex, int lastTypeIndex, IList<ICollection<PathNode>> pathNodes, double extraLength) {
            if (lastTypeIndex < firstTypeIndex) {
                return new List<PathNode>(2) { firstNode, lastNode };
            }
            int midIndex = firstTypeIndex + (lastTypeIndex - firstTypeIndex) / 2;
            var paths = new List<ThreeNodePath>(pathNodes[midIndex].Count);
            foreach (var node in pathNodes[midIndex]) {
                var path = new ThreeNodePath(firstNode, node, lastNode);
                if (path.Length + extraLength < _bestPathLength)
                    paths.Add(path);
            }

            if (paths.Count == 0) {
                return null;
            }

            paths.Sort();

            List<PathNode> bestFirstPath = null;
            List<PathNode> bestLastPath = null;

            {
                var path = paths[0];
                var firstPath = FindGoodPath(firstNode, path.MidNode, firstTypeIndex, midIndex - 1, pathNodes, path.LastSegmentLength);
                var lastPath = FindGoodPath(path.MidNode, lastNode, midIndex + 1, lastTypeIndex, pathNodes, path.FirstSegmentLength);


                var newLength = PathLength(firstPath) + PathLength(lastPath);
                bestFirstPath = firstPath;
                bestLastPath = lastPath;


            }
            if (bestFirstPath == null) {
                return null;
            }

            var result = new List<PathNode>(bestFirstPath.Count + bestLastPath.Count - 1);
            result.AddRange(bestFirstPath);
            result.AddRange(bestLastPath.GetRange(1, bestLastPath.Count - 1));
            return result;
        }

        private double PathLength(List<PathNode> path) {
            var enumerator = path.GetEnumerator();
            enumerator.MoveNext();
            var a = enumerator.Current;
            if (!enumerator.MoveNext()) return 0;
            var b = enumerator.Current;
            double result = DirectPath.length(a, b);
            while (enumerator.MoveNext()) {
                a = b;
                b = enumerator.Current;
                result += DirectPath.length(a, b);
            }
            return result;
        }
    }
}
